{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Lecture 17:  Classes II\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "These lecture notes are partly adapted from https://www.python-course.eu/."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Data Hiding via Attribute Types\n",
    "\n",
    "When we create instance variables of a class, we must decide what level of access \"users\" of the class should have over the data and procedural attributes.  Some OOP languages enforce these distinctions,  Python uses a special naming convention to \"signal the attribute type.\" \n",
    "\n",
    "* Private (prefixed with `__` or `_`): these attributes should only be used inside of the class definition.  Among these, attributes with `__`  are strictly private and essentially invisible from outside. Attributes with `_` can be used from outside but really should not be.  \n",
    "* Public (no underscore prefix): these attributes can and should be freely used.\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "class TestingAttributes():\n",
    "    __slots__ = ['__val', '_val', 'val']\n",
    "    def __init__(self):\n",
    "        self.__val = \"I am strictly private.\"\n",
    "        self._val = \"I am private but accessible from outside.\"\n",
    "        self.val = \"I am public.\""
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "a = TestingAttributes()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "ename": "AttributeError",
     "evalue": "'TestingAttributes' object has no attribute '__val'",
     "output_type": "error",
     "traceback": [
      "\u001b[0;31m---------------------------------------------------------------------------\u001b[0m",
      "\u001b[0;31mAttributeError\u001b[0m                            Traceback (most recent call last)",
      "\u001b[0;32m<ipython-input-3-3e19e2bd1a2b>\u001b[0m in \u001b[0;36m<module>\u001b[0;34m\u001b[0m\n\u001b[0;32m----> 1\u001b[0;31m \u001b[0ma\u001b[0m\u001b[0;34m.\u001b[0m\u001b[0m__val\u001b[0m\u001b[0;34m\u001b[0m\u001b[0;34m\u001b[0m\u001b[0m\n\u001b[0m",
      "\u001b[0;31mAttributeError\u001b[0m: 'TestingAttributes' object has no attribute '__val'"
     ]
    }
   ],
   "source": [
    "a.__val"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "'I am private but accessible from outside.'"
      ]
     },
     "execution_count": 4,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "a._val"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "'I am public.'"
      ]
     },
     "execution_count": 5,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "a.val"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Print Representation of a Class: `__str__`\n",
    "\n",
    "If we print an object of our class `coordinate`, notice that the output is not very human readable.  We can create user--defined way of a \"pretty\" object description using the special `__str__` method. \n",
    "\n",
    "Python calls the `__str__` method when you print a class object.  We can choose what it does!  Suppose we want to it to print `<3, 4>` when called on an object with x-coordinate `3` and y-coordinate `4`.  Let us write a `__str__` method to do this."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [],
   "source": [
    "class A:\n",
    "    \"\"\"Test printing of objects.\"\"\"\n",
    "    pass"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [],
   "source": [
    "a = A()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "<__main__.A object at 0x10db77f90>\n"
     ]
    }
   ],
   "source": [
    "print(a)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [],
   "source": [
    "class A:\n",
    "    \"\"\"Test printing of objects.\"\"\"\n",
    "    def __str__(self):\n",
    "        return 'print representation'"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [],
   "source": [
    "a = A()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "print representation\n"
     ]
    }
   ],
   "source": [
    "print(a)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Name:\n",
    "    \"\"\"Class to represent a person's name.\"\"\"\n",
    "    __slots__ = ['_f', '_m', '_l']\n",
    "    \n",
    "    def __init__(self, first, last, middle=''):\n",
    "        self._f = first\n",
    "        self._m = middle\n",
    "        self._l = last\n",
    "    \n",
    "    def __str__(self):\n",
    "        if len(self._m):\n",
    "            return '{}. {}. {}'.format(self._f[0], self._m[0], self._l) \n",
    "        return '{}. {}'.format(self._f[0], self._l)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [],
   "source": [
    "n1 = Name('Shikha', 'Singh')\n",
    "n2 = Name('Iris', 'Howley', 'K.')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "S. Singh\n",
      "I. K. Howley\n"
     ]
    }
   ],
   "source": [
    "print(n1)\n",
    "print(n2)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## OOP Principle: Encapsulation\n",
    "\n",
    "Encapsulation is the bundling of data with the methods that operate on that data. It is often accomplished by providing two kinds of procedural attributes: \n",
    "* methods for retrieving or accessing the values of attributes, called **getter methods** or accessor methods. Getter methods do not change the values of attributes, they just return the values, and \n",
    "* methods used for changing the values of attributes are called **setter methods**."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Coordinate(object):\n",
    "    \"\"\"Represents the coordinates of a point.\"\"\"\n",
    "    __slots__ = ['_x', '_y']\n",
    "    \n",
    "    def __init__(self, x, y):\n",
    "        self._x = x\n",
    "        self._y = y\n",
    "        \n",
    "    def getX(self):\n",
    "        return self._x\n",
    "    \n",
    "    def getY(self):\n",
    "        return self._y\n",
    "        \n",
    "    def __str__(self):\n",
    "        return '<{},{}>'.format(self._x, self._y)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {},
   "outputs": [],
   "source": [
    "p = Coordinate(4, 3)\n",
    "q = Coordinate(0, 2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "<4,3>\n"
     ]
    }
   ],
   "source": [
    "print(p)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "<0,2>\n"
     ]
    }
   ],
   "source": [
    "print(q)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 19,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "4"
      ]
     },
     "execution_count": 19,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p.getX()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "3"
      ]
     },
     "execution_count": 20,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p.getY()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 21,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "4"
      ]
     },
     "execution_count": 21,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p.getX() + q.getX()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 22,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "5"
      ]
     },
     "execution_count": 22,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p.getY() + q.getY()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**Note.** Explicilty calling getter methods as functions, especially if they are used a lot, can be cumbersome.  It would be nice if we can write `p.x + p.y` instead of `p.getX() + q.getX()`."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## `@property`\n",
    "\n",
    "**Annotations `@`.**   Python provides a rich collection of syntactic notes that can change how code is interpreted, called annotations.  These are typically prefixed with the at-sign (@).\n",
    "\n",
    "* Accessor methods do not change the state of the calling object and are used just to retrieve some information about the object\n",
    "* `@property` annotation (Treat a procedural attribute as a data attribute):  If we’d like to treat an accessor method as-if it were a data attribute, we can use the @property annotation\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Coordinate(object):\n",
    "    \"\"\"Represents the coordinates of a point.\"\"\"\n",
    "    slots = ['_x', '_y']\n",
    "    def __init__(self, x, y):\n",
    "        self._x = x\n",
    "        self._y = y\n",
    "    \n",
    "    @property\n",
    "    def x(self):\n",
    "        return self._x\n",
    "    \n",
    "    @property\n",
    "    def y(self):\n",
    "        return self._y"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "metadata": {},
   "outputs": [],
   "source": [
    "p = Coordinate(4, 3)\n",
    "q = Coordinate(0, 2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 25,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "4"
      ]
     },
     "execution_count": 25,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p.x # notice no parenthesis!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 26,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "3"
      ]
     },
     "execution_count": 26,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p.y"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "(4, 5)"
      ]
     },
     "execution_count": 27,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p.x + q.x, p.y + q.y"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Example: Name Class\n",
    "\n",
    "In the previous example, we used `@property` to access the data attributes of the Coordinate objects.  This is not the only use of the annotation, we can use it for any accessor method that we want to invoke as if it was a data attribute."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 28,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Name:\n",
    "    \"\"\"Class to represent a person's name.\"\"\"\n",
    "    __slots__ = ['_f', '_m', '_l']\n",
    "    \n",
    "    def __init__(self, first, last, middle=''):\n",
    "        self._f = first\n",
    "        self._m = middle\n",
    "        self._l = last\n",
    "    \n",
    "    def initials(self):\n",
    "        if len(self._m):\n",
    "            return '{}. {}. {}.'.format(self._f[0], self._m[0], self._l[0]).upper()\n",
    "        return '{}. {}.'.format(self._f[0], self._l[0]).upper()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 29,
   "metadata": {},
   "outputs": [],
   "source": [
    "me = Name('Shikha', 'Singh')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 30,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "'S. S.'"
      ]
     },
     "execution_count": 30,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "me.initials()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 31,
   "metadata": {},
   "outputs": [],
   "source": [
    "iris = Name('Iris', 'Howley', 'K')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 32,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "'I. K. H.'"
      ]
     },
     "execution_count": 32,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "iris.initials()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 33,
   "metadata": {
    "scrolled": true
   },
   "outputs": [
    {
     "data": {
      "text/plain": [
       "<bound method Name.initials of <__main__.Name object at 0x10db4e870>>"
      ]
     },
     "execution_count": 33,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "iris.initials"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**Using @property.** Let us annotate the initials class with `@property` to make treat initials as if it is a data attribute of `Name` class. \n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 34,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Name:\n",
    "    \"\"\"Class to represent a person's name.\"\"\"\n",
    "    __slots__ = ['_f', '_m', '_l']\n",
    "    \n",
    "    def __init__(self, first, last, middle=''):\n",
    "        self._f = first\n",
    "        self._m = middle\n",
    "        self._l = last\n",
    "    \n",
    "    @property\n",
    "    def initials(self):\n",
    "        if len(self._m):\n",
    "            return '{}. {}. {}.'.format(self._f[0], self._m[0], self._l[0]).upper()\n",
    "        else:\n",
    "            return '{}. {}.'.format(self._f[0], self._l[0]).upper()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 35,
   "metadata": {},
   "outputs": [],
   "source": [
    "iris = Name('Iris', 'Howley', 'K')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 36,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "'I. K. H.'"
      ]
     },
     "execution_count": 36,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "iris.initials  # notice no parenthesis!"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Building the Coordinate Class\n",
    "\n",
    "Let us apply the concept we have learn to finish up our coordinate class.  \n",
    "\n",
    "We have already implemented:\n",
    "* `x(self)`: returns the x coordinate of the calling instance (annotated with @)\n",
    "* `y(self)`: returns the y coordinate of the calling instance (annotated with @)\n",
    "* `__str__`: print representation of the class\n",
    "\n",
    "We will now implement:\n",
    "* `_subX(self, other)`:  subtracts the x coordinates of self and other and returns the answer\n",
    "* `_subY(self, other)`:  subtracts the y coordinate of self and other and returns the answer\n",
    "* `dist(self, others)`: computes the Euclidean distance between coordinates self and other\n",
    "* `radius(self)`: computes and returns the radius of the calling coordindate (annotated with @)\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "#### Distance between Two Points\n",
    "\n",
    "Let us define a method `distance` for the `Coordinate` class which when called on an object of the class computes the Euclidean distance between that object and another point.  Recall that the Eucliden distance between two points $p_1 = (x_1, y_1)$ and $p_2 = (x_2, y_2)$ is\n",
    "\n",
    "$$ d(p_1, p_2) = \\sqrt{((x_1-x_2)^2 + (y_1-y_2)^2}.$$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 37,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Coordinate(object):\n",
    "    \"\"\"Represents the coordinates of a point.\"\"\"\n",
    "    __slots__ = ['_x', '_y']\n",
    "    def __init__(self, x, y):\n",
    "        self._x = x\n",
    "        self._y = y\n",
    "        \n",
    "    @property\n",
    "    def x(self):\n",
    "        return self._x\n",
    "    \n",
    "    @property\n",
    "    def y(self):\n",
    "        return self._y\n",
    "    \n",
    "    def _subX(self, other):\n",
    "        \"\"\"Subtracts the x coordinates of self \n",
    "        and other and returns the answer\"\"\"\n",
    "        return self._x - other._x\n",
    "    \n",
    "    def _subY(self, other):\n",
    "        \"\"\"Subtracts the y coordinates of self \n",
    "        and other and returns the answer\"\"\"\n",
    "        return self._y - other._y\n",
    "    \n",
    "    def dist(self, other):\n",
    "        sqX = self._subX(other)**2\n",
    "        sqY = self._subY(other)**2\n",
    "        return round((sqX + sqY)**0.5, 2)\n",
    "    \n",
    "    @property\n",
    "    def radius(self):\n",
    "        \"\"\"Returns the distance of the point from (0,0)\"\"\"\n",
    "        origin = Coordinate(0,0)\n",
    "        return self.dist(origin)\n",
    "    \n",
    "    def __str__(self):\n",
    "        return '<{}, {}>'.format(self._x, self._y)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 38,
   "metadata": {},
   "outputs": [],
   "source": [
    "p1 = Coordinate(0,2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 39,
   "metadata": {},
   "outputs": [],
   "source": [
    "p2 = Coordinate(3, 4)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 40,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "0"
      ]
     },
     "execution_count": 40,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p1.x"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 41,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "4"
      ]
     },
     "execution_count": 41,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p2.y"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 42,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "3.61"
      ]
     },
     "execution_count": 42,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p1.dist(p2)  # call distance on object pt, zero is argument"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 43,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "3.61"
      ]
     },
     "execution_count": 43,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "round(13**0.5,2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 44,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "2.0"
      ]
     },
     "execution_count": 44,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p1.radius"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 45,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "5.0"
      ]
     },
     "execution_count": 45,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p2.radius"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Summary\n",
    "\n",
    "Today we saw how Python supports data abstraction (separating the data and details of the implementation from the user) via :\n",
    "\n",
    "* Data hiding: via attribute naming conventions (private, public)\n",
    "* Encapsulation: bundling together of data and methods that provide an interface to the data \n",
    "    * accessor methods that can be simply implemented using `@property` annotation"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.4"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
